
//#include "tasandid_stdafx.h"
#include "mySemaphore.h"
#include "myException.h"



///TODO: see
/// http://www.boost.org/doc/html/threads.html




mySemaphore::mySemaphore(char *cName, long initialCount, long maximumCount, int accessMode, BOOL canInherit)
{
	std::string sName(cName);

	init_inner(sName, initialCount, maximumCount, accessMode, canInherit);
};

mySemaphore::mySemaphore(std::string &sName, long initialCount, long maximumCount, int accessMode, BOOL canInherit)
{																	   
	init_inner(sName, initialCount, maximumCount, accessMode, canInherit);
};


void mySemaphore::init_inner(std::string &sName, long initialCount, long maximumCount, int accessMode, BOOL canInherit)
{
	/// initialization
	sa = NULL;
	semaphore = NULL;
	semaphoreName = "";
	initCount = 0;
	maxCount  = 0;
	inherit   = TRUE;

	/// create it
	if (accessMode == CREATE)
    {
		/// security handling
		sa = (LPSECURITY_ATTRIBUTES)HeapAlloc(GetProcessHeap(), 0, sizeof(SECURITY_ATTRIBUTES));
        sa->nLength = sizeof(SECURITY_ATTRIBUTES);
        sa->lpSecurityDescriptor = NULL; ///no restrictions
        sa->bInheritHandle = canInherit;

		/// make sure the count is legal
		/// normally, you want to create the semaphore will an initialcount = 0;
		/// therefore, the original state of the semaphore is locked.
		/// once the system finishes all the initialization work, it can
		/// call unlock() to increase it to be 1, so other threads can
		/// start to use it
		if (maximumCount < initialCount) 
		{
			maximumCount = 1;
			initialCount = 0;
		}

		/// create the semaphore
        if ((semaphore = CreateSemaphore(sa, initialCount, maximumCount, sName.c_str())) == NULL)
        {
			int errCode = 0;
			std::string header = "CreateSemaphore failed";
			getErrorDetail(errCode, header);
			myException excp(errCode, header);
            throw excp;
            return;
        }

		/// save the key values
		initCount = initialCount;
		maxCount  = maximumCount;
		inherit   = canInherit;

		/// save the name, if specified
		if (sName.compare("")) 
			semaphoreName.append(sName);

	}

	/// open it
    else
    {
        if ((semaphore = OpenSemaphore(SEMAPHORE_ALL_ACCESS, canInherit, sName.c_str())) == NULL)
        {
			int errCode;
			std::string header = "OpenSemaphore failed";
			getErrorDetail(errCode, header);
            myException excp(errCode, header);
			throw excp;
            return;
        }
    }
}

mySemaphore::~mySemaphore()
{
	if (sa) HeapFree(GetProcessHeap(), 0, sa);
	if (semaphore) closeSemaphore();
}

void mySemaphore::getErrorDetail(int& errCode, std::string& header)
{
	char buf[256];
	LPVOID lpDetailMsgBuf;
	UINT errorCode;

    FormatMessage
    (
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		NULL, 
		errorCode = GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR)&lpDetailMsgBuf, 
		sizeof(buf),				//0,					//roland
		NULL
	);

	memset(buf, 0, sizeof(buf));
	wsprintf(buf, ": %s", (LPTSTR)lpDetailMsgBuf);
    LocalFree(lpDetailMsgBuf);	
	
	header.append(std::string(buf));
	errCode = errorCode;
}

void mySemaphore::closeSemaphore()
{
    if (!CloseHandle(semaphore))
    {
		int errCode = 0;
		std::string header = "CloseHandle failed";
		getErrorDetail(errCode, header);
        myException excp(errCode, header);
		throw excp;
    }
	semaphore = NULL;
}

BOOL mySemaphore::lock(BOOL wait, DWORD milliseconds)
{
	if (wait == TRUE)
	{
		if (WaitForSingleObject(semaphore, INFINITE) == WAIT_FAILED)
		{
			int errCode = 0;
			std::string header = "WaitForSingleObject failed";
			getErrorDetail(errCode, header);
			myException excp(errCode, header);
			throw excp;
		}
		return TRUE;
	}
	else 
	{
		DWORD retVal = WaitForSingleObject(semaphore, milliseconds);

		if (retVal == WAIT_TIMEOUT) 
		{
			#pragma warning (push)		
			#pragma warning (disable : 4127)		///warning C4127: conditional expression is constant		
			MYASSERT(false);
			#pragma warning (pop)			
			
			return FALSE;				 /// Semaphore was nonsignaled, so a time-out occurred.
		}
		
		if (retVal == WAIT_OBJECT_0) 
			return TRUE;				 /// The semaphore object was signaled.

		if (retVal == WAIT_FAILED)
		{
			int errCode = 0;
			std::string header = "WaitForSingleObject failed";
			getErrorDetail(errCode, header);
			myException excp(errCode, header);
			throw excp;
		}
	}

	return FALSE;
}

void mySemaphore::unlock()
{
	if (!ReleaseSemaphore(semaphore, 1, NULL))
	{
		int errCode = 0;
		std::string header = "ReleaseSemaphore failed: ";
		header.append(semaphoreName);
		getErrorDetail(errCode, header);
		myException excp(errCode, header);
		throw excp;
	}
}

/*
long mySemaphore::getSemaphoreCount()
{
	long currCount = 0;

	if ( !ReleaseSemaphore(semaphore,0,&currCount) )
	{
		int errCode = 0;
		string header = "ReleaseSemaphore failed";
		getErrorDetail(errCode,header);
		myException excp(errCode,header);
		throw excp;
	}
    return currCount;
}
*/






///mySemaphore *debug_semaphore = NULL;


/*
void semaphore_init()
{
	if (debug_semaphore == NULL)
	{
		std::string name = std::string("DEBUG_SEMAPHORE");
		debug_semaphore = new mySemaphore(name, 1, 1, CREATE, TRUE);	
	}
};
*/




///TODO: see http://en.wikipedia.org/wiki/Singleton_pattern
///TODO: see http://en.wikipedia.org/wiki/Double-checked_locking



bool free_indexes_list_inited = false;
__declspec(thread) static int semaphores_taken_for_this_thread_a[256];		///declspec(thread) can be used only with static data

//TODO - memory of these two will not be released by code at program shutdown
//std::list<int> *free_indexes_list = NULL;					 
//mySemaphore *inner_lock = NULL; ///("myFlexiSemaphore_inner_lock", 1, 1, CREATE, TRUE);
std::list<int> *free_indexes_list = NULL;			///you cant rely on the initialization order of global variables, so we must use pointers...
mySemaphore *outer_lock = NULL; ///("myFlexiSemaphore_outer_lock", 1, 1, CREATE, TRUE);	 ///you cant rely on the initialization order of global variables


///####################################################################### 

void __cdecl free_indexes_list_freemem_at_exit()
{							 
	delete free_indexes_list;	
	delete outer_lock;
	
	free_indexes_list	= NULL;
	outer_lock			= NULL;
};

///####################################################################### 

myFlexiSemaphore::myFlexiSemaphore()
: mySemaphore("myFlexiSemaphore_instance", 1, 1, CREATE, TRUE)
{
	///MYASSERT(myindex == -1);
	
	///inner lock 2 is not necessary because program starts with only one thread and first myFlexiSemaphore is inited already at the start
	/*
	std::string lock2_name("myFlexiSemaphore_inner_lock2");
	mySemaphore inner_lock2(lock2_name, 1, 1, CREATE, TRUE);
	inner_lock2.lock();
	*/
	
	/*
	if (inner_lock == NULL)
	{
		inner_lock = new mySemaphore("myFlexiSemaphore_inner_lock", 1, 1, CREATE, TRUE);		///we must use new, because we cant be sure that inner_lock object is initalized already when current object is being initalized
		free_indexes_list = new std::list<int>;
	}
	*/
	
	
	//inner_lock2.unlock();
	
	
	
	if (outer_lock == NULL)															///we assume that during startup we have only one thread
	{
		outer_lock = new mySemaphore("myFlexiSemaphore_outer_lock", 1, 1, CREATE, TRUE);
		free_indexes_list = new std::list<int>;
		
		
		atexit(free_indexes_list_freemem_at_exit);
	}
	
	

	if (!free_indexes_list_inited)				///we assume that during startup we have only one thread
	{
		free_indexes_list_inited = true;
		
		for (int i = 255; i >= 0; i--)
		{
			free_indexes_list->push_back(i);		
		}		
	}
	
	
	
	
	
	outer_lock->lock();
	
	

	_ASSERT(free_indexes_list->size() > 0);
	

	///semaphores_taken_for_this_thread = 0;
	//std::list<int>::const_iterator myindex_iter = free_indexes_list.back();
	//const int &myindex_in = myindex_iter;	
	const int &myindex_in = free_indexes_list->back();
	myindex = myindex_in;
	free_indexes_list->pop_back();
	
	//inner_lock = new mySemaphore("myFlexiSemaphore_inner_lock", 1, 1, CREATE, TRUE);
	
	
	///semaphores_taken_for_this_thread_a[myindex] = 0;
	
	
	
	outer_lock->unlock();
	
};

///####################################################################### 

myFlexiSemaphore::~myFlexiSemaphore() 
{
	_ASSERT(myindex != -1);


	outer_lock->lock();
	///free_indexes_list->push_back(myindex);		cant push_back because we cant be sure that previous user-object freed counters in all of its threads
	
	myindex = -1;
	//delete inner_lock;
	
/*	
	std::list<int>::iterator iter;
	//for (int i = 255; i >= 0; i--)
	for (iter = free_indexes_list.begin(); iter != free_indexes_list.end(); iter++)
	{
		//free_indexes_list->push_back(i);	
			
	}
*/	
	
	outer_lock->unlock();
	
	
	/*
	if (myindex == 0)		///the first and last one
	{
	
	
	}
	*/
};

///####################################################################### 
	
void myFlexiSemaphore::lock()
{
	_ASSERT(myindex != -1);


	//#ifndef USE_OPENMP
	//inner_lock->lock();	   ///seems to be necessary, because increment-decrement are NOT atomic
	//#endif


#pragma warning (push)
#pragma warning (disable : 6385)						 ///warning C6385: Invalid data: accessing 'semaphores_taken_for_this_thread_a', the readable size is '1024' bytes, but '-4' bytes might be read
	if (semaphores_taken_for_this_thread_a[myindex] == 0)	
	{
		//#pragma omp atomic
		semaphores_taken_for_this_thread_a[myindex]++;		///it is thread-specific variable, so the operation does not have to be atomic
#pragma warning (pop)		
		
		//#ifndef USE_OPENMP		
		//inner_lock->unlock();
		//#endif		
		
		mySemaphore::lock();	   ///must be outside of inner lock, or deadlock occurs
	}
	else
	{	
		//#pragma omp atomic
		semaphores_taken_for_this_thread_a[myindex]++;
		
		
		//#ifndef USE_OPENMP		
		//inner_lock->unlock();
		//#endif		
	}
	
	
	
};

///#######################################################################	

void myFlexiSemaphore::unlock()
{	 	
	_ASSERT(myindex != -1);
	
	
	//#ifndef USE_OPENMP	
	//inner_lock->lock();		  ///seems to be necessary, because increment-decrement are NOT atomic
	//#endif
	

	//#pragma omp atomic

#pragma warning (push)
#pragma warning (disable : 6385)						 ///warning C6385: Invalid data: accessing 'semaphores_taken_for_this_thread_a', the readable size is '1024' bytes, but '-4' bytes might be read
	semaphores_taken_for_this_thread_a[myindex]--;		///it is thread-specific variable, so the operation does not have to be atomic
	
	
	if (semaphores_taken_for_this_thread_a[myindex] == 0)
	{
#pragma warning (pop)	
		mySemaphore::unlock();
	}
	
	
	//#ifndef USE_OPENMP			
	//inner_lock->unlock();
	//#endif
}; 

#if 0
int myFlexiSemaphore::get_semaphores_taken_for_this_thread() C
{
	return semaphores_taken_for_this_thread_a[myindex];
};
#endif





myFlexiSemaphore startup_init_guaranteer();

